#!/bin/bash

set -bm

CONFIG="config"
PORTMIN="4000"
PORTMAX="4100"
PREFIX=$(dirname $0)
TMP="/tmp"
CONNECTIONHELPER="/tmp/connection-helper"
DEBUG="NO"
GRIDCONFIG="/tmp/mcaparams.conf"
PIDFILE="/tmp/qcg.pid"
MACHINEFILE=$TMP"/machinefile"

debug() {
# default: ignore display-only commands (echo and cat), and silently run other commands
  if [ "$1" != "echo" -o "$1" != "cat" ]; then
     $* 1>/dev/null 2>/dev/null
     return $?
  fi
  return 0
}

NBPIDS=0
declare -a PIDS

# HOSTNAME must be defined.
# Usually it is, but let's be paranoid
if (( ${#HOSTNAME} == 0 )) ; then
        export HOSTNAME=`hostname`;
fi
DOMAIN=${HOSTNAME#*.}
if ((${#DOMAIN} == ${#HOSTNAME}))  ; then
	HOSTNAME=$(hostname).$(dnsdomainname) ;
else
	HOSTNAME=$(hostname);
fi

help_exit() {
    [ -n "$1" ] && echo "\`$1': option not allowed" >&2
    echo "Usage: $0 [OPTIONS]" >&2
    echo "  where options are:" >&2
    echo "  -c|--conf           /path/to/configfile  defaults to \`$CONFIG'" >&2
    echo "  -m|--port-min       min port to use (all servers on all machines) defaults to $PORTMIN" >&2
    echo "  -M|--port-max       max port to use (all servers on all machines) defaults to $PORTMAX" >&2
    echo "  -p|--prefix         /path/to/all/binaries defaults to \`$PREFIX'" >&2
    echo "  -t|--tmp            /path/to/tmp/directory defaults to \`$TMP'" >&2
    echo "  -h|--help           displays this help and exits" >&2
    echo "  -d|--debug          will display all services stdout and stderr at the end of run" >&2
    exit 1
}

TEMP=$(getopt -o c:m:M:p:t:hd --long conf:,port-min:,port-max:,prefix:,tmp:,help,debug -n "qcgrte-deployer" -- "$@")

if [ $? != 0 ]; then
    help_exit "getopt failure"
fi

while true; do
    case "$1" in
	-c|--conf)
	    CONFIG=$2
	    shift 2
	    ;;
        -d|--debug)
debug() {
# if the command is display-only, then run it to stderr
   if [ "$1" == "echo" -o "$1" == "cat" ]; then
      $* >&2
      return $?
   else
# else, display the command on stderr then run it to stderr
      echo "(qcgrte-deployer debug)" $* >&2
      $* >&2
      return $?
   fi
}
            DEBUG="YES"
            shift
            ;;
	-m|--port-min)
	    PORTMIN=$2
	    shift 2
	    ;;
	-M|--port-max)
	    PORTMAX=$2
	    shift 2
	    ;;
	-p|--prefix)
	    PREFIX=$2
	    shift 2
	    ;;
	-t|--tmp)
	    TMP=$2
	    shift 2
	    ;;
	-h|--help)
	    help_exit ""
	    ;;	
	*)
	    [ -z "$1" ] && break
	    help_exit $1
	    ;;
    esac
done

if [ ! -f $CONFIG ]; then
    echo "config file not found" >&2
    help_exit ""
fi

testexec() {
    if [ ! -x $PREFIX/$1 ]; then
	echo "$PREFIX/$1 does not exists or is not executable" >&2
	exit 1
    fi
}

for binary in mpiexec broker connectionhelper proxy frontale; do
    testexec $binary
done

runcmd() {
    outputfile=$1
    shift

    runbg() {
	parent=$1
	shift
	debug echo "qcgrte-deployer now executes [$*]." 	
	$* 1>$outputfile.out 2>$outputfile.err &
	CHILD=$!
	trap "debug kill -9 $CHILD" USR2
	wait $CHILD
#	[ $? -ne 0 ] && debug echo "process $* ($CHILD) did not terminate correctly" >&2
#	debug echo "$CHILD ($?) ($*) is done" 
#	debug kill -USR1 $parent
	exit 0
    }

    runbg $$ $* &
    child=$!
    PIDS[$((NBPIDS++))]=$child    
    echo $child >> $PIDFILE
}

trap "debug terminate" USR1

BROKERPORT=$((PORTMIN++))
FRONTALPORT=$((PORTMIN++))
PROXYPORT=$((PORTMIN++))
PROXYSERVICEPORT=$((PORTMIN++))
HELPERPORT=$((PORTMIN++))
CONNHELPERPORT=$((PORTMIN++))
PROXYPORTMIN=$((PORTMIN++))
PROXYPORTMAX=$((PORTMAX))

if [ ! -f $PIDFILE ]
then
    touch $PIDFILE
else 
    rm $PIDFILE
    touch $PIDFILE
fi


if [ $PORTMIN -gt $PORTMAX ]; then
    echo "unable to allocate all ports in the specified range" >&2
    echo "  (portmax should at least be equal to $PORTMIN)" >&2
    exit 1
fi

runcmd $TMP/broker $PREFIX/broker -c $CONFIG -p $BROKERPORT -f $FRONTALPORT &
echo $PREFIX "/broker -c" $CONFIG "-p" $BROKERPORT "-f" $FRONTALPORT
sleep 60

FRONTALES=$(awk ' $1 ~ /^[0-9]+$/ {print $4}' $CONFIG)
i=0
for f in $FRONTALES; do    
#    echo  "starting frontale on " $f
    runcmd $TMP/frontale-$i ssh $f $PREFIX/frontale -g $BROKERPORT -s $FRONTALPORT  -c $HELPERPORT -S $HOSTNAME  &
    i=$((i+1))
echo "ssh" $f $PERFIX "/frontale -g" $BROKERPORT  "-s" $FRONTALPORT "-c" $HELPERPORT "-S" $HOSTNAME
done

sleep 120
echo $PREFIX "/proxy -n -u"  $PROXYSERVICEPORT "-p" $PROXYPORT "-s" $BROKERPORT "-S" $HOSTNAME  
runcmd $TMP/proxy $PREFIX/proxy -u $PROXYSERVICEPORT -p $PROXYPORT -s $BROKERPORT -S $HOSTNAME & 

sleep 20
i=0
for f in $FRONTALES; do
    MACHINES=$(awk '$3 != "" { output=($4 == "'$f'") } { if(output && $3 == "") print}' $CONFIG)
    for m in $MACHINES; do
#	echo "conn helper on " $m
	runcmd $TMP/machine-$i ssh $m $PREFIX/connectionhelper -F $f -f $HELPERPORT -g $CONNHELPERPORT&
	i=$((i+1))
    done
done

MCAPARAMFILE=~/.openmpi/mca-params.conf
MCAPARAMS=$(mktemp -t mcaparams.$HOSTNAME.XXXXXXXXX)

BROKERMCAPARAM="oob_tcp_connection_helper"
#CONHELPMCAPARAM="oob_tcp_connhelper"
BROKERMCAPARAM1="btl_tcp_connection_helper"
#CONHELPMCAPARAM1="btl_tcp_connhelper"

OOBPORTMINMCAPARAM="oob_tcp_port_min"
OOBPORTMAXMCAPARAM="oob_tcp_port_max"
BTLPORTMINMCAPARAM="btl_tcp_port_min"
BTLPORTMAXMCAPARAM="btl_tcp_port_max"

MACHINES=$(awk ' $1 !~ /^[0-9]+$/ {print $1}' $CONFIG)
MYMACHINEFILE=$(mktemp -t machinefile.XXXXXXXXXX)
FRONTALES=$(awk ' $1 ~ /^[0-9]+$/ {print $4}' $CONFIG)
NP=0
for f in $FRONTALES; do    
    MYPORTMIN=$(grep $f $CONFIG | head -n 1 | awk {'print $7'})
    MYPORTMAX=$(grep $f $CONFIG | head -n 1 | awk {'print $8'})
#    echo $MYPORTMIN " and " $MYPORTMAX
    MYMCAPARAMS=$(mktemp -t mcaparams.$f.XXXXXXXXX)
    echo $BROKERMCAPARAM" = 127.0.0.1:"$CONNHELPERPORT >> $MYMCAPARAMS
#    echo $CONHELPMCAPARAM" = "$CONNECTIONHELPER  >> $MYMCAPARAMS
    echo $BROKERMCAPARAM1" = 127.0.0.1:"$CONNHELPERPORT >> $MYMCAPARAMS
#    echo $CONHELPMCAPARAM1" = "$CONNECTIONHELPER  >> $MYMCAPARAMS
    echo $OOBPORTMINMCAPARAM" = "$MYPORTMIN  >> $MYMCAPARAMS
    echo $OOBPORTMAXMCAPARAM" = "$MYPORTMAX  >> $MYMCAPARAMS
    echo $BTLPORTMINMCAPARAM" = "$MYPORTMIN  >> $MYMCAPARAMS
    echo $BTLPORTMAXMCAPARAM" = "$MYPORTMAX  >> $MYMCAPARAMS
    debug echo "MCA PARAMETRES FILE" 
    debug echo "*****************************" 
    debug cat $MYMCAPARAMS
    debug echo "*******************************" 
    MACHINES=$(awk '$3 != "" { output=($4 == "'$f'") } { if(output && $3 == "") print}' $CONFIG)
    for m in $MACHINES; do
	if [ $m != $HOSTNAME ]
	then 
	    scp $MYMCAPARAMS $m:$GRIDCONFIG
	    ssh $m chmod +r $GRIDCONFIG
	    echo $m >> $MYMACHINEFILE
	    NP=$((NP+1))
	else
	    cp $MYMCAPARAMS $GRIDCONFIG
	    chmod +r $GRIDCONFIG
	fi
    done
    cp $MYMACHINEFILE $MACHINEFILE
done

for f in $FRONTALES; do    
    MACHINES=$(awk '$3 != "" { output=($4 == "'$f'") } { if(output && $3 == "") print}' $CONFIG)
    for m in $MACHINES; do
	if [ $m != $HOSTNAME ]
	then 
	    scp $MACHINEFILE $m:$MACHINEFILE
	    ssh $m chmod +r $MACHINEFILE
	else
	    chmod +r $MACHINEFILE
	fi
    done
done
